@@ -57,10 +57,14 @@ namespace Gyoto {
  */
 #define GYOTO_WORLDLINE_PROPERTIES(c)					\
   Gyoto::Property("Gyoto::Worldline", "Time-like or null geodesic."),	\
+    GYOTO_PROPERTY_BOOL(c, Integ31, Integ4D, _integ31,			\
+			"Whether to integrate the 3+1 or 4D equation of geodesics.") \
     GYOTO_PROPERTY_BOOL(c, HighOrderImages, PrimaryOnly, _secondary,	\
 			"Whether to stop Photon integration at 180° deflection.") \
-    GYOTO_PROPERTY_BOOL(c, ParallelTransport, NoParallelTransport, _parallelTransport,	\
+    GYOTO_PROPERTY_BOOL(c, ParallelTransport, NoParallelTransport, _parallelTransport, \
 			"Whether to perform parallel transport of a local triad (used for polarization).") \
+    GYOTO_PROPERTY_DOUBLE(c, MaxCrossEqplane, _maxCrossEqplane,		\
+			  "Maximum number of crossings of the equatorial plane allowed for this worldline") \
     GYOTO_PROPERTY_DOUBLE(c, RelTol, _relTol,				\
 			  "Relative tolerance for the adaptive step integrators.") \
     GYOTO_PROPERTY_DOUBLE(c, AbsTol, _absTol,				\
@@ -100,12 +104,16 @@ namespace Gyoto {
    GYOTO_WORLDLINE macro.
 */
 #define GYOTO_WORLDLINE_ACCESSORS(c)					\
+  void c::_integ31(bool s) {integ31(s);}		                \
+  bool c::_integ31() const {return integ31();}                          \
   void c::_secondary(bool s) {secondary(s);}				\
   bool c::_secondary() const {return secondary();}			\
   void c::_parallelTransport(bool s) {parallelTransport(s);}		\
   bool c::_parallelTransport() const {return parallelTransport();}	\
   void c::_adaptive(bool s) {adaptive(s);}				\
   bool c::_adaptive() const {return adaptive();}			\
+  void c::_maxCrossEqplane(double max){maxCrossEqplane(max);}	      	\
+  double c::_maxCrossEqplane()const{return maxCrossEqplane();}	     	\
   void c::_relTol(double f){relTol(f);}					\
   double c::_relTol()const{return relTol();}				\
   void c::_absTol(double f){absTol(f);}					\
@@ -166,6 +174,8 @@ namespace Gyoto {
   bool _adaptive () const ;				\
   void _secondary (bool sec) ;				\
   bool _secondary () const ;				\
+  void _integ31 (bool sec) ;			\
+  bool _integ31 () const ;			\
   void _parallelTransport (bool sec) ;			\
   bool _parallelTransport () const ;			\
   void _maxiter (size_t miter) ;			\
@@ -176,6 +186,8 @@ namespace Gyoto {
   void _deltaMin(double h1);				\
   void _absTol(double);					\
   double _absTol()const;				\
+  void _maxCrossEqplane(double);       			\
+  double _maxCrossEqplane()const;			\
   void _relTol(double);					\
   double _relTol()const;				\
   void _deltaMax(double h1);				\
@@ -346,6 +358,14 @@ class Gyoto::Worldline
    */
   double reltol_;
 
+  /**
+   * \brief Maximum number of crossings of equatorial plane
+   *
+   * Used to determine how much higher-order image features
+   * should be kept.
+   */
+  double maxCrossEqplane_;
+
   // Constructors - Destructor
   // -------------------------
  public: 
@@ -464,6 +484,19 @@ class Gyoto::Worldline
    */
   std::string integrator() const ;
 
+  /**
+   * \brief Set the integrator kind to 3+1 or 4D
+   *
+   * Initialize #state_ to use the required geodesic equation.
+   *
+   */
+  void integ31(bool integ);
+
+  /**
+   * \brief Get the kind of geodesic equation integrated by #state_
+   */
+  bool integ31() const ;
+
   /**
    * \brief Get #delta_min_
    */
@@ -485,6 +518,9 @@ class Gyoto::Worldline
   void relTol(double); ///< Set #reltol_
   double relTol()const; ///< Get #reltol_
 
+  void maxCrossEqplane(double); ///< Set #maxCrosEqplane_
+  double maxCrossEqplane()const; ///< Get #maxCrossEqplane_
+
   /**
    * Get delta max at a given position
    *
@@ -578,7 +614,12 @@ class Gyoto::Worldline
 
   /**
    * Return pointer to array holding the previously set
-   * Metric-specific constants of motion
+   * Metric-specific constants of motion.
+   *
+   * This function returns a pointer to the actual storage location
+   * and should be handled with care. std::vector<double>
+   * Worldline:constantsOfMotion() const provides a convenient way to
+   * retrieve a copy of the content.
    */
   double const * getCst() const ; ///< Returns the worldline's cst of motion (if any)
 
@@ -586,9 +627,30 @@ class Gyoto::Worldline
   /**
    * The will (re)allocate Worldline::cst_, copy cst into it, and set
    * Worldline::cst_n_.
+   *
+   * This is the same as void
+   * Worldline:constantsOfMotion(std::vector<double> const cstv) using
+   * a C-style array instead of a vector.
    */
   void setCst(double const * cst, size_t const ncsts) ;
 
+  /// Set Metric-specific constants of motion
+  /**
+   * The will (re)allocate Worldline::cst_, copy cst into it, and set
+   * Worldline::cst_n_.
+   *
+   * This is the same as getCst using a vector instead of a C-style array.
+   */
+  void constantsOfMotion(std::vector<double> const cstv) ;
+
+  /// Return a copy of the Metric-specific constants of motion
+  /**
+   * This funtion return a copy of the constants of motion. getCst()
+   * can be used to retrieve a pointer to the actual array used
+   * internally which is slightly more efficient for read-only access.
+   */
+  std::vector<double> constantsOfMotion() const ;
+
   /// Set or re-set the initial condition prior to integration.
   /**
    * \param gg    Gyoto::SmartPointer to the Gyoto::Metric in this universe;
@@ -815,6 +877,8 @@ class Gyoto::Worldline::IntegState::Generic : public SmartPointee {
    * when calling nestStep().
    */
   Worldline * line_;
+  // add 31 flag
+  bool integ_31_; ///< Whether to integrate the 3+1 equation of geodesics instead of the 4D one.
   double delta_; ///< Integration step (current in case of #adaptive_).
   bool adaptive_; ///< Whether to use an adaptive step
   bool parallel_transport_; ///< Whether to parallel-transport base vectors
@@ -879,6 +943,12 @@ class Gyoto::Worldline::IntegState::Generic : public SmartPointee {
    */
   virtual std::string kind()=0;
 
+  /**
+   * \brief Defines the kind of geodesic equation to integrate (3+1, 4D)
+   */
+  void integ31(bool integ);
+  bool integ31() const ;
+
   /// Make one step.
   /**
    * \param[out] coord Next position-velocity;
